package htmling;

use strict;

### CONSTANTS
$htmling::dblspace = "  ";
$htmling::indentspace = $htmling::dblspace x 2;
$htmling::headerspace = $htmling::indentspace;
$htmling::comment_indent = $htmling::indentspace x 2;

### PUBLIC GLOBALS
$htmling::comments_type = "smart";

### GLOBALS
$htmling::htmlfile = "";
$htmling::indent = 0;

sub do_module {
   my ($module) = @_;

   $htmling::htmlfile = "$module->{'name'}.html";
   print "Generating $htmling::htmlfile...\n";
   open (OUT, ">$htmling::htmlfile");

   print OUT "<HTML>\n";
   print OUT "<HEAD>\n";
   print OUT "   <TITLE> Fortran 90 module $module->{'name'} </TITLE>\n";
   print OUT "</HEAD>\n";
   print OUT "<H1> Module $module->{'name'} </H1>\n";
   print OUT "<PRE>module $module->{'name'}\n";

   list_uses (@{$module->{'uses'}});
   list_html ("Types", map (($_->{'type'} eq "type" ? ($_) : ()), @{$module->{'ocontains'}}));
   list_html ("Variables", map (($_->{'type'} eq "var" ? ($_) : ()), @{$module->{'ocontains'}}));
   list_html ("Interfaces", map (($_->{'type'} eq "interface" ? ($_) : ()), @{$module->{'ocontains'}}));
   list_html ("Subroutines and functions", map (($_->{'type'} eq "subroutine" || $_->{'type'} eq "function" ? ($_) : ()), @{$module->{'ocontains'}}));

   print OUT "\nend module $module->{'name'}\n";
   do_comments ($module->{'comments'}, 1);

   my (@list);
   @list = map (($_->{'type'} eq "type" ? ($_) : ()), @{$module->{'ocontains'}});
   print OUT "\n<HR><H2> Description of Types </H2>\n" if @list;
   do_html (@list);
   @list = map (($_->{'type'} eq "var" ? ($_) : ()), @{$module->{'ocontains'}});
   print OUT "\n<HR><H2> Description of Variables </H2>\n" if @list;
   do_html (@list);
   @list = map (($_->{'type'} eq "interface" ? ($_) : ()), @{$module->{'ocontains'}});
   print OUT "\n<HR><H2> Description of Interfaces </H2>\n" if @list;
   do_html (@list);
   @list = map (($_->{'type'} eq "subroutine" || $_->{'type'} eq "function" ? ($_) : ()), @{$module->{'ocontains'}});
   print OUT "\n<HR><H2> Description of Subroutines and Functions </H2>\n" if @list;
   do_html (@list);
   
   print OUT "</HTML>\n";
   close OUT;
}

sub list_uses {
   if (@_) {
      print OUT "\n${htmling::indentspace}${htmling::headerspace}! Uses\n";
      my ($use);
      foreach $use (@_) {
         print OUT "${htmling::indentspace}use <A HREF=\"$use.html\">$use</A>\n";
      }
   }
}

sub list_html {
   my ($title) = shift;

   if (@_) {
      print OUT "\n${htmling::indentspace}${htmling::headerspace}! $title\n";
      my ($struct);
      foreach $struct (@_) {
         my ($href) = "<A HREF=\"${htmling::htmlfile}#$struct->{'type'}_" .
            lc ($struct->{'name'}) . "\">$struct->{'name'}</A>";
         print OUT $htmling::indentspace;
         if ($struct->{'type'} eq "var") {
            print OUT var2str ($struct, $href) . "\n";
         } else {
            print OUT "$struct->{'vis'} $struct->{'type'} $href";
            if ($struct->{'type'} eq "subroutine" || $struct->{'type'} eq "function") {
               print OUT " (" . join (", ", @{$struct->{'parms'}}) . ")";
            }
            print OUT "\n";
         }
      }
   }
}

sub do_html {
   if (@_) {
      my ($struct);

      foreach $struct (@_) {
         my ($vis);
         if (exists $struct->{'vis'}) {
            $vis = $struct->{'vis'} . " ";
         } else {
            $vis = "";
         }

         if (! $htmling::indent) {
            print OUT "<A NAME=\"$struct->{'type'}_" . lc ($struct->{'name'}) .
               "\"><H3>$struct->{'name'}</H3></A>\n";
            print OUT "<PRE>";
         }

         print OUT $htmling::indentspace x $htmling::indent;
         if ($struct->{'type'} eq "var") {
             print OUT var2str ($struct) . "\n";
         } elsif ($struct->{'type'} eq "mprocedure") {
             die "do_html: bare module procedure $struct->{'name'} (no enclosing module)"
                 unless exists $struct->{'bind'};
             print OUT
                 "module procedure <A HREF=\"#$struct->{'bind'}->{'type'}_" .
                 lc ($struct->{'name'}) . "\">$struct->{'name'}</A>\n";
         } elsif ($struct->{type} eq "subroutine" ||
                  $struct->{type} eq "function") {
             print OUT $vis;
             print OUT typing::type_to_f90 ($struct->{rtype}) . " "
                 if exists $struct->{rtype};
             print OUT "recursive " if $struct->{recursive};
             print OUT "$struct->{'type'} $struct->{'name'}";
             print OUT " (" . join (", ", @{$struct->{parms}}) . ")";
             print OUT " result $struct->{result}" if exists $struct->{result};
             print OUT "\n";
         } else {
             print OUT "${vis}$struct->{'type'} $struct->{'name'}\n";
         }

         $htmling::indent++;

         if ($struct->{'type'} eq "var" || $struct->{'type'} eq "mprocedure") {
         } elsif ($struct->{'type'} eq "type" || $struct->{'type'} eq "interface") {
             do_html (@{$struct->{'ocontains'}});
         } elsif ($struct->{'type'} eq "subroutine" || $struct->{'type'} eq "function") {
             my @interest = @{$struct->{'parms'}};
             push @interest, $struct->{result} if exists $struct->{result};
             my $name;
             foreach $name (@interest) {
                 my (@things) = values %{$struct->{contains}->{lc $name}};
                 die "Confused by/no declaration for parameter $name of $struct->{'type'} $struct->{'name'}" if scalar @things != 1;
                 my ($var) = $things[0];
                 print OUT $htmling::indentspace x $htmling::indent . var2str ($var) . "\n";
             }
         } else {
             die "do: I don't know what a $struct->{'type'} is";
         }

         $htmling::indent--;

         if ($struct->{'type'} ne "var" && $struct->{'type'} ne "mprocedure") {
            print OUT $htmling::indentspace x $htmling::indent . "end $struct->{'type'} $struct->{'name'}\n";
         }

         do_comments ($struct->{'comments'}, ! $htmling::indent);
      }
   }
}

# Pass comments and a flag saying if you want to end the current <PRE> block.
sub do_comments {
   my ($comments, $endpre) = @_;
   if ($comments eq "") {
      print OUT "</PRE>\n" if $endpre;
      return;
   }

   #print OUT "\n" unless $htmling::indent;

   if ($htmling::comments_type eq "preformatted") {
      my ($s) = $htmling::indentspace x $htmling::indent . $htmling::comment_indent;
      $comments =~ s/^/$s/m if $htmling::indent;
      $comments =~ s/^\n*//s;
      $comments =~ s/\n*$//s;
      print OUT $comments, "\n";
      print OUT "</PRE>\n" if $endpre;
   } else {
      print OUT "</PRE>\n";
      print OUT "<DL><DD><DL><DD>\n" if $htmling::indent;
      if ($htmling::comments_type eq "html") {
      } elsif ($htmling::comments_type eq "smart") {
         my @newcomments = ();
         my $verbmode = 0;
         my @listmode = ();
         my $line;
         foreach $line (split ("\n", $comments)) {
            if ($verbmode) {
               if ($line =~ /^>/) {
                  warn "`$line' found while already in verbatim mode";
                  substr ($line, 0, 1) = " ";
                  push @newcomments, $line;
               } elsif ($line =~ /^</) {
                  $verbmode = 0;
                  substr ($line, 0, 1) = " ";
                  push @newcomments, $line . "</PRE>";
               } elsif ($line =~ /^v/) {
                  warn "`$line' found while already in verbatim mode";
                  substr ($line, 0, 1) = " ";
                  push @newcomments, $line;
               } else {
                  push @newcomments, $line;
               }
               next;
            }

            while ($line =~ /(\A|\W)_(\w|\w.*?\w)_(\Z|\W)/) {
                my ($left, $mid, $right) = ("$`$1<I>", $2, "</I>$3$'");
                $mid =~ s/_/ /g;
                $line = $left . $mid . $right;
            }
            while ($line =~ /(\A|\W)\*(\w|\w.*?\w)\*(\Z|\W)/) {
                my ($left, $mid, $right) = ("$`$1<STRONG>", $2, "</STRONG>$3$'");
                $mid =~ s/\*/ /g;
                $line = $left . $mid . $right;
            }

            if ($line =~ /^( *)-/) {
               if (@listmode) {
                  if ($listmode[$#listmode] > length ($1)) {
                     push @listmode, length ($1);
                     push @newcomments, $1 . "<UL>";
                  } else {
                     while ($listmode[$#listmode] != length ($1)) {
                        push @newcomments, " " x $listmode[$#listmode] . "</UL>";
                        pop @listmode;
                        die "Unindented to invalid position in `$line'" unless @listmode;
                     }
                  }
               } else {
                  push @listmode, length ($1);
                  push @newcomments, $1 . "<UL>";
               }
               push @newcomments, $1 . "<LI> " . substr ($line, length ($&));
            } elsif ($line =~ /^>/) {
               warn "Verbatim mode started in list mode" if @listmode;
               $verbmode = 1;
               substr ($line, 0, 1) = " ";
               push @newcomments, "<PRE>" . $line;
            # Ignore $line =~ /^</ because it may be an HTML tag.
            } elsif ($line =~ /^v/) {
                warn "Verbatim mode started in list mode" if @listmode;
                $verbmode = 1;
                substr ($line, 0, 1) = " ";
                push @newcomments, "<PRE>" . $line;
            } elsif ($line =~ /^\s*$/) {
                push @newcomments, "<P>";
            } elsif (@listmode) {
               $line =~ /^( *)/;
               while (@listmode && $listmode[$#listmode] > length ($1)) {
                  push @newcomments, " " x $listmode[$#listmode] . "</UL>";
                  pop @listmode;
               }
               push @newcomments, $line;
            } else {
               push @newcomments, $line;
            }
         }
         my $list;
         foreach $list (@listmode) {
             push @newcomments, " " x $list . "</UL>";
         }
         $comments = join ("\n", @newcomments);
      } else {
         die "Unsupported comments type `$htmling::comments_type'";
      }
      $comments =~ s/<P>\n(<P>\n)+/<P>\n/g;
      $comments =~ s/<P>\n$//;
      $comments =~ s/^<P>\n//;
      $comments =~ s/<P>/<DD>/g if $htmling::indent;
      print OUT $comments . "\n";
      print OUT "</DL></DL>\n" if $htmling::indent;
      print OUT "<PRE>" unless $endpre;
   }
}

sub var2str {
   my ($var, $href) = @_;

   my ($attribs) = join (", ", (exists $var->{'vis'} ? $var->{'vis'} : ()), @{$var->{'tempattribs'}});
   $attribs = ", $attribs" if $attribs;
   my ($typestr) = typing::type_to_f90 ($var->{'vartype'});
   my ($initial) = (exists $var->{'initial'} ? " = " . typing::expr_to_f90 ($var->{'initial'}) : "");
   $href = $var->{'name'} unless $href;
   return "$typestr$attribs :: $href$initial";
}

1;
